import datetime
import os
import queue
import sys
import threading
import time
import atexit

from kuai_log._datetime import aware_now


class Stream:
    @classmethod
    def stdout(cls, msg):
        print(msg)

    @classmethod
    def print(cls, *args, sep=' ', end='\n', file=None, flush=True, sys_getframe_n=2):
        args = (str(arg) for arg in args)  # REMIND 防止是数字不能被join
        msg = sep.join(args) + end
        print(msg)

class BulkStream:
    q = queue.SimpleQueue()
    _lock = threading.Lock()
    _has_start_bulk_stdout = False

    @classmethod
    def _bulk_real_stdout(cls):
        with cls._lock:
            msg_str_all = ''
            while not cls.q.empty():
                msg_str_all += str(cls.q.get()) + '\n'
            if msg_str_all:
                sys.stdout.write(msg_str_all)

    @classmethod
    def stdout(cls, msg):
        with cls._lock:
            cls.q.put(msg)

    @classmethod
    def print(cls, *args, sep=' ', end='\n', file=None, flush=True, sys_getframe_n=2):
        args = (str(arg) for arg in args)  # REMIND 防止是数字不能被join
        msg = sep.join(args) + end
        fra = sys._getframe(1)  # noqa
        line = fra.f_lineno
        full_path = fra.f_code.co_filename  # type :str
        # file_name = full_path.split('/')[-1].split('\\')[-1]
        full_path_and_line = f'"{full_path}:{line}"'
        fun = fra.f_code.co_name
        full_msg = f'{aware_now()} - {full_path_and_line} - {fun} - [print] - {msg}'
        color_msg = f'\033[0;34m{full_msg}\033[0m'
        with cls._lock:
            cls.q.put(color_msg)

    @classmethod
    def _when_exit(cls):
        # stdout_raw('结束 stdout_raw')
        return cls._bulk_real_stdout()

    @classmethod
    def start_bulk_stdout(cls):
        def _bulk_stdout():
            while 1:
                cls._bulk_real_stdout()
                time.sleep(0.1)

        if not cls._has_start_bulk_stdout:
            cls._has_start_bulk_write = True
            threading.Thread(target=_bulk_stdout, daemon=True).start()

    @classmethod
    def patch_print(cls):
        """
        Python有几个namespace，分别是

        locals

        globals

        builtin

        其中定义在函数内声明的变量属于locals，而模块内定义的函数属于globals。


        https://codeday.me/bug/20180929/266673.html   python – 为什么__builtins__既是模块又是dict

        :return:
        """
        try:
            __builtins__.print = cls.print
        except AttributeError:
            """
            <class 'AttributeError'>
            'dict' object has no attribute 'print'
            """
            # noinspection PyUnresolvedReferences
            __builtins__['print'] = cls.print
        # traceback.print_exception = print_exception  # file类型为 <_io.StringIO object at 0x00000264F2F065E8> 单独判断，解决了，不要加这个。

OsStream = Stream


if os.name == 'nt':  # windows io性能差
    OsStream = BulkStream
    BulkStream.start_bulk_stdout()
    BulkStream.patch_print()
    atexit.register(BulkStream._when_exit)




import re
from calendar import day_abbr, day_name, month_abbr, month_name
from datetime import datetime as datetime_
from datetime import timedelta, timezone
from time import localtime, strftime

tokens = r"H{1,2}|h{1,2}|m{1,2}|s{1,2}|S{1,6}|YYYY|YY|M{1,4}|D{1,4}|Z{1,2}|zz|A|X|x|E|Q|dddd|ddd|d"

pattern = re.compile(r"(?:{0})|\[(?:{0}|!UTC)\]".format(tokens))


class datetime(datetime_):  # noqa: N801
    def __format__(self, spec):
        if spec.endswith("!UTC"):
            dt = self.astimezone(timezone.utc)
            spec = spec[:-4]
        else:
            dt = self

        if not spec:
            spec = "%Y-%m-%dT%H:%M:%S.%f%z"

        if "%" in spec:
            return datetime_.__format__(dt, spec)

        year, month, day, hour, minute, second, weekday, yearday, _ = dt.timetuple()
        microsecond = dt.microsecond
        timestamp = dt.timestamp()
        tzinfo = dt.tzinfo or timezone(timedelta(seconds=0))
        offset = tzinfo.utcoffset(dt).total_seconds()
        sign = ("-", "+")[offset >= 0]
        h, m = divmod(abs(offset // 60), 60)

        rep = {
            "YYYY": "%04d" % year,
            "YY": "%02d" % (year % 100),
            "Q": "%d" % ((month - 1) // 3 + 1),
            "MMMM": month_name[month],
            "MMM": month_abbr[month],
            "MM": "%02d" % month,
            "M": "%d" % month,
            "DDDD": "%03d" % yearday,
            "DDD": "%d" % yearday,
            "DD": "%02d" % day,
            "D": "%d" % day,
            "dddd": day_name[weekday],
            "ddd": day_abbr[weekday],
            "d": "%d" % weekday,
            "E": "%d" % (weekday + 1),
            "HH": "%02d" % hour,
            "H": "%d" % hour,
            "hh": "%02d" % ((hour - 1) % 12 + 1),
            "h": "%d" % ((hour - 1) % 12 + 1),
            "mm": "%02d" % minute,
            "m": "%d" % minute,
            "ss": "%02d" % second,
            "s": "%d" % second,
            "S": "%d" % (microsecond // 100000),
            "SS": "%02d" % (microsecond // 10000),
            "SSS": "%03d" % (microsecond // 1000),
            "SSSS": "%04d" % (microsecond // 100),
            "SSSSS": "%05d" % (microsecond // 10),
            "SSSSSS": "%06d" % microsecond,
            "A": ("AM", "PM")[hour // 12],
            "Z": "%s%02d:%02d" % (sign, h, m),
            "ZZ": "%s%02d%02d" % (sign, h, m),
            "zz": tzinfo.tzname(dt) or "",
            "X": "%d" % timestamp,
            "x": "%d" % (int(timestamp) * 1000000 + microsecond),
        }

        def get(m):
            try:
                return rep[m.group(0)]
            except KeyError:
                return m.group(0)[1:-1]

        return pattern.sub(get, spec)


def aware_now():
    now = datetime_.now()
    timestamp = now.timestamp()
    local = localtime(timestamp)

    try:
        seconds = local.tm_gmtoff
        zone = local.tm_zone
    except AttributeError:
        offset = datetime_.fromtimestamp(timestamp) - datetime_.utcfromtimestamp(timestamp)
        seconds = offset.total_seconds()
        zone = strftime("%Z")

    tzinfo = timezone(timedelta(seconds=seconds), zone)

    return datetime.combine(now.date(), now.time().replace(tzinfo=tzinfo))
